import 'dart:convert';

import 'binary_search/binary_search.dart';

String assignStatePos(String variable, List<(int, int)> ranges, bool negate) {
  if (ranges.isEmpty) {
    throw ArgumentError('Must not be empty', 'ranges');
  }

  if (negate) {
    return 'state.pos += $variable > 0xffff ? 2 : 1';
  }

  final has16Bit = ranges.any((e) => e.$1 <= 0xffff || e.$2 <= 0xffff);
  final has32Bit = ranges.any((e) => e.$1 > 0xffff || e.$2 > 0xffff);
  return switch ((has16Bit, has32Bit)) {
    (false, false) => 'state.pos += $variable > 0xffff ? 2 : 1',
    (false, true) => 'state.pos += 2',
    (true, false) => 'state.pos++',
    (true, true) => 'state.pos += $variable > 0xffff ? 2 : 1',
  };
}

String charAt(List<(int, int)> ranges, bool negate) {
  if (is32BitRanges(ranges, negate)) {
    return 'runeAt';
  }

  return 'codeUnitAt';
}

String escapeString(String text, [bool quote = true]) {
  text = text.replaceAll('\\', r'\\');
  text = text.replaceAll('\b', r'\b');
  text = text.replaceAll('\f', r'\f');
  text = text.replaceAll('\n', r'\n');
  text = text.replaceAll('\r', r'\r');
  text = text.replaceAll('\t', r'\t');
  text = text.replaceAll('\v', r'\v');
  text = text.replaceAll('\'', '\\\'');
  text = text.replaceAll('\$', r'\$');
  if (!quote) {
    return text;
  }

  return '\'$text\'';
}

({String name, String size}) getCharReader(bool is16Bit, String name) {
  if (is16Bit) {
    return (name: 'codeUnitAt', size: '1');
  }

  return (name: 'runeAt', size: '$name > 0xffff ? 2 : 1');
}

(int, int) getUnicodeRange() {
  return const (0, 0x10ffff);
}

bool is32BitRanges(List<(int, int)> ranges, bool negate) {
  if (ranges.isEmpty) {
    throw ArgumentError('Must not be empty', 'ranges');
  }

  if (negate) {
    return true;
  }

  return ranges.any((e) => e.$1 > 0xffff || e.$2 > 0xffff);
}

bool isTypeNullable(String? type) {
  if (type == null) {
    return true;
  }

  type = type.trim();
  return type.endsWith('?') || type == 'dynamic';
}

List<(int, int)> normalizeRanges(List<(int, int)> ranges) {
  final binarySearch = BinarySearch();
  final result = binarySearch.normalizeRanges(ranges);
  return result;
}

String rangesToPredicate(String name, List<(int, int)> ranges, bool negate) {
  final binarySearch = BinarySearch();
  final result = binarySearch.generatePredicate(name, ranges, negate);
  return result;
}

String render(String template, Map<String, String> values,
    {bool removeEmptyLines = true}) {
  for (final entry in values.entries) {
    final key = entry.key;
    final value = entry.value;
    template = template.replaceAll('{{$key}}', value);
  }

  final buffer = StringBuffer();
  final lines = const LineSplitter().convert(template);
  for (var i = 0; i < lines.length; i++) {
    final line = lines[i];
    if (line.trim().isEmpty && removeEmptyLines) {
      continue;
    }

    if (i < lines.length - 1) {
      buffer.writeln(line);
    } else {
      buffer.write(line);
    }
  }

  return buffer.toString();
}

String testLiteral({
  required List<int> codeUnits,
  required String end,
  required String input,
  String? start,
}) {
  if (codeUnits.isEmpty) {
    throw ArgumentError('Must not be empty', 'string');
  }

  final buffer = StringBuffer();
  buffer.write('state.pos');
  if (codeUnits.length - 1 > 0) {
    buffer.write(' + ${codeUnits.length - 1}');
  }

  buffer.write(' < $end &&');
  for (var i = 0; i < codeUnits.length; i++) {
    final char = codeUnits[i];
    var statePos = start == null ? 'state.pos' : 'state.pos - $start';
    if (i > 0) {
      statePos = '$statePos + $i';
    }

    buffer.write('$input.codeUnitAt($statePos) == $char');
    if (i < codeUnits.length - 1) {
      buffer.writeln(' &&');
    }
  }

  return buffer.toString();
}

class _Code extends _Expression {
  final String code;

  const _Code(this.code);

  @override
  String toString() {
    return code;
  }
}

abstract class _Expression {
  const _Expression();
}

class _Group extends _Expression {
  final _Expression expression;

  _Group(this.expression);

  @override
  String toString() {
    return '($expression)';
  }
}

class _Triple extends _Code {
  _Triple(super.code);
}
